/*******************************************************************************

   DLLMAIN.C

   Authors: Peter Loan
            Krystyne Blaikie

   Version 6.0, June 8, 2011

   Copyright (c) 2004-2011 MusculoGraphics, a division of Motion Analysis Corp.
   All rights reserved.

   Description: This file contains the main routine for the DLL version of
      a dynamics simulation (for running within SIMM), and several other
      utility functions that are independent of the musculoskeletal model
      in the simulation. This file is generated by SIMM when you choose the
      "project files" option in the "Save Dynamics" dialog box. You will
      probably want to generate it once (not every time you change your
      model) and then customize it to your specific simulation.

*******************************************************************************/
#include "universal.h"

/*************** DEFINES (for this file only) *********************************/


/*************** STATIC GLOBAL VARIABLES (for this file only) *****************/
char sim_version[] = "6.0";
SystemInfo si;
dpDataSetup dataSetup;
dpDataFrame dataFrame;
int (*dataCallbackFunc)(dpSimmModelID, int, void*) = NULL;
double t, *y, *dy, *work;
FILE* kin_out = NULL;
char* output_kinetics_file = NULL;
dpMessageStruct mess;
dpModelInfo* gModelInfo = NULL;
dpBoolean isStandAlone = dpNo;
char sd_error_msg[CHARBUFFER];

/**************** GLOBAL VARIABLES (used in only a few files) *****************/
dpModelStruct* sdm;
MotionData* input_data = NULL;
dpSimulationParameters params;
char current_dir[CHARBUFFER];

/*************** EXTERNED VARIABLES (declared in another file) ****************/
extern char buffer[];
extern dpBoolean verbose;

/*************** PROTOTYPES for STATIC FUNCTIONS (for this file only) *********/



DLL int dpGetVersion(void)
{
   return 600; // version 6.0
}


DLL void* dpGetModelInfo(void)
{
   if (gModelInfo)
   {
      return (void*)gModelInfo;
   }
   else
   {
      int i, info[50], slider[6];

      sdinfo(info);

      gModelInfo = (dpModelInfo*)simm_calloc(1, sizeof(dpModelInfo));

      gModelInfo->num_segments = info[1]+1;
      gModelInfo->num_closed_loops = info[4];
      gModelInfo->num_tree_joints = info[1];
      gModelInfo->num_joints = info[1] + info[4]; // nbod + nloop
      gModelInfo->joint_type = (dpJointType*)simm_calloc(gModelInfo->num_joints, sizeof(dpJointType));
      gModelInfo->contactEnabled = dpNo;
#if CONTACT_DETECTION
      gModelInfo->contactEnabled = dpYes;
#endif


      for (i=0; i<gModelInfo->num_joints; i++)
      {
         sdjnt(i, info, slider);
         switch(info[0])
         {
            case 1:
               gModelInfo->joint_type[i] = dpPin;
               break;
            case 2:
               gModelInfo->joint_type[i] = dpUniversal;
               break;
            case 3:
               gModelInfo->joint_type[i] = dpGimbal;
               break;
            case 4:
               gModelInfo->joint_type[i] = dpBall;
               break;
            case 5:
               gModelInfo->joint_type[i] = dpSlider;
               break;
            case 6:
               gModelInfo->joint_type[i] = dpFree;
               break;
            case 7:
               gModelInfo->joint_type[i] = dpCylindrical;
               break;
            case 8:
               gModelInfo->joint_type[i] = dpPlanar;
               break;
            case 9:
               gModelInfo->joint_type[i] = dpWeld;
               break;
            case 10:
               gModelInfo->joint_type[i] = dpBushing;
               break;
            case 11:
               gModelInfo->joint_type[i] = dpBearing;
               break;
            case 20:
               gModelInfo->joint_type[i] = dpReversePlanar;
               break;
            case 21:
               gModelInfo->joint_type[i] = dpReverseFree;
               break;
            case 22:
               gModelInfo->joint_type[i] = dpReverseBushing;
               break;
            case 23:
               gModelInfo->joint_type[i] = dpReverseBearing;
               break;
            default:
               gModelInfo->joint_type[i] = dpUnknownJoint;
               break;
         }
      }
   }

   return (void*)gModelInfo;
}


/* dpInitSimulation: this function is called when the "initialize" button
 * is pressed in the Dynamics Tool, before dpSetModel() is called.
 */
DLL int dpInitSimulation(dpSimulationParameters* _params)
{
   copyParameters(_params, &params);

   chdir(params.workingDir);

   memset(&dataSetup, 0, sizeof(dpDataSetup));
   memset(&dataFrame, 0, sizeof(dpDataFrame));

   return 0;
}


/* init_dll_model: this function is called by dpSetModel to finish
 * initializing the sdm structure after the basic model data has been
 * filled in.
 */
void init_dll_model(void)
{
   int i, j, k, info[50], total_joints;
   double current_mass, current_inertia[3][3], current_vector[3];
   ReturnCode rc;

   sdinfo(info);
   sdm->num_body_segments = info[1] + 1;    // include ground
   sdm->nq = info[2] + info[7];
   sdm->nu = info[2];
   sdm->num_closed_loops = info[4];
   total_joints = info[1] + info[4]; // nbod + nloop
   sdm->num_constraints = info[3];
   sdm->num_user_constraints = info[10];

   // Make room for the loop joints.
   if (total_joints > sdm->num_joints)
   {
      sdm->joint = (dpJointStruct*)simm_realloc(sdm->joint, total_joints * sizeof(dpJointStruct), &rc);
      sdm->num_joints = total_joints;
   }

   init_joints(dpNo);

   // Set the gravity vector.
   sdgrav(sdm->gravity);

   // Compute the coefficients for the constraint functions, restraint functions 
   // and moving muscle point functions in the model.
   for (i = 0; i < sdm->num_functions; i++)
      calc_function_coefficients(sdm->function[i]);

#if CONTACT_DETECTION
   // Determine how to get from one segment to another.
   makepaths();
#endif

   // Set the masses, inertia matrices, and joint vectors in the SD/FAST code,
   // based on their values in the model passed in from SIMM. But only try
   // to set them if the new values are different than the current ones,
   // because the question marks may have been removed from model.sd, which
   // means that you cannot change these values at runtime.
   for (i = 1; i < sdm->num_body_segments; i++)
   {
      sdgetmass(i - 1, &current_mass);
      if (NOT_EQUAL_WITHIN_ERROR(sdm->body_segment[i].mass, current_mass))
      {
         sdmass(i - 1, sdm->body_segment[i].mass);
         if (check_for_sderror("SET_MASS") == 19)
            sdm->body_segment[i].mass = current_mass;
      }

      sdgetiner(i - 1, current_inertia);
      for (j = 0; j < 3; j++)
      {
         for (k = 0; k < 3; k++)
         {
            if (NOT_EQUAL_WITHIN_ERROR(sdm->body_segment[i].inertia[j][k], current_inertia[j][k]))
            {
               sdiner(i - 1, sdm->body_segment[i].inertia);
               if (check_for_sderror("SET_INERTIA") == 19)
                  sdgetiner(i - 1, sdm->body_segment[i].inertia);
               break;
            }
         }
      }

      sdgetbtj(i - 1, current_vector);
      if (NOT_EQUAL_WITHIN_ERROR(sdm->body_segment[i].body_to_joint[0], current_vector[0]) ||
          NOT_EQUAL_WITHIN_ERROR(sdm->body_segment[i].body_to_joint[1], current_vector[1]) ||
          NOT_EQUAL_WITHIN_ERROR(sdm->body_segment[i].body_to_joint[2], current_vector[2]))
      {
         sdbtj(i - 1, sdm->body_segment[i].body_to_joint);
         if (check_for_sderror("SET_BODY_TO_JOINT") == 19)
            sdgetbtj(i - 1, sdm->body_segment[i].body_to_joint);
      }

      sdgetitj(i - 1, current_vector);
      if (NOT_EQUAL_WITHIN_ERROR(sdm->body_segment[i].inboard_to_joint[0], current_vector[0]) ||
          NOT_EQUAL_WITHIN_ERROR(sdm->body_segment[i].inboard_to_joint[1], current_vector[1]) ||
          NOT_EQUAL_WITHIN_ERROR(sdm->body_segment[i].inboard_to_joint[2], current_vector[2]))
      {
         sditj(i - 1, sdm->body_segment[i].inboard_to_joint);
         if (check_for_sderror("SET_INBOARD_TO_JOINT") == 19)
            sdgetitj(i - 1, sdm->body_segment[i].inboard_to_joint);
      }
   }

   // Make space for the muscle moment arms, if appropriate.
   if ((params.outputOptions.muscleMomentArms_Muscles.num_items > 0 && params.outputOptions.muscleMomentArms_Gencoords.num_items > 0) ||
      (params.outputOptions.muscleMoments_Muscles.num_items > 0 && params.outputOptions.muscleMoments_Gencoords.num_items > 0) ||
      params.outputOptions.totalMuscleMoments.num_items > 0 ||
      (params.outputOptions.optimizedActivations_Muscles.num_items > 0 && params.outputOptions.optimizedActivations_Gencoords.num_items > 0))
   {
      for (i = 0; i < sdm->num_muscles; i++)
      {
         sdm->muscles[i].nummomentarms = sdm->nu;
         sdm->muscles[i].momentarms = (double*)simm_malloc(sdm->muscles[i].nummomentarms * sizeof(double));
      }
   }
}


/* dpRegisterCallback: this function stores the name of a callback function in
 * a global variable. This callback is called with output data after each iteration
 * of the integrator. It is used to pass simulation results and messages back
 * into SIMM.
 */
DLL int dpRegisterCallback(int (*dataCallback)(dpSimmModelID, int, void*))
{
   dataCallbackFunc = dataCallback;
   return 0;
}


/* dpSetModel: this function is called when the "initialize" button is pressed
 * in the Dynamics Tool. It takes a dpModelStruct containing a complete description
 * of the SIMM model, and copies the information to the global variable sdm. It
 * then assembles the model, initializes the state vector and other variables,
 * then opens the appropriate output files.
 */
DLL int dpSetModel(dpModelStruct* model)
{
   int i, status, num_muscle_states;
   double param[1] = {-1.0};
   char *dllparameters_file_name = "dllparams.txt";
   char *contactparameters_file_name = "dllcontactparams.txt";

   // Copy the dpModelStruct to memory that is local to the DLL.
   sdm = (dpModelStruct*)simm_calloc(1, sizeof(dpModelStruct));
   copyModel(model, sdm);

   // Check to make sure the model is consistent with the one built into the DLL (from model.sd).
   if (checkModel(sdm) < 0)
   {
      mess.id = dpFinished;
      if (dataCallbackFunc)
         dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
      return -1;
   }

   // Count the number of muscle states, and quit the simulation
   // if there are fewer than zero (meaning that there was an
   // error counting the states).
   num_muscle_states = count_muscle_states(sdm);
   if (num_muscle_states < 0)
      return -1;

   sdm->neq = sdm->nu + sdm->nq + num_muscle_states;

   y = (double*)simm_malloc(sdm->neq * sizeof(double));
   dy = (double*)simm_malloc(sdm->neq * sizeof(double));
   work = (double*)simm_malloc(6 * sdm->neq * sizeof(double));

   if (params.inputMotionFile)
      set_up_input_data(params.inputMotionFile, &input_data);

   read_dllparameters_file(dllparameters_file_name, sdm);

#if CONTACT_DETECTION
   if (read_contact_parameters_file(contactparameters_file_name) == code_bad)
      return -1;
#endif

   init_dll_model();

   // Check to make sure the output options have been specified properly,
   // and find indices for the names.
   if (checkOutputOptions() < 0)
   {
      mess.id = dpFinished;
      sprintf(mess.text, "Cannot run this simulation DLL with %s.", model->name);
      if (dataCallbackFunc)
         dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
      return -1;
   }

   if (set_initial_conditions(&t, y, dy, input_data) != code_fine)
   {
      mess.id = dpFinished;
      sprintf(mess.text, "Unable to initialize the dynamic simulation with input data.");
      if (dataCallbackFunc)
         dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
      return -1;
   }

   init_motion();

   sdinit();
   check_for_sderror("SDINIT");
   
   // Enable constraint stabilization.
   sdstab(2.0 * params.baumgarte_stabilization, params.baumgarte_stabilization * params.baumgarte_stabilization);

   if (sdm->newInverseSimulation == dpYes)
   {
      sdm->newInverseSimulation = dpNo;
      if (assemble_model(t, y) == code_bad)
      {
         mess.id = dpFinished;
         sprintf(mess.text, "Unable to initialize the dynamic simulation.");
         if (dataCallbackFunc)
            dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
         return -1;
      }
      sdm->newInverseSimulation = dpYes;
   }
   else
   {
      if (assemble_model(t, y) == code_bad)
      {
         mess.id = dpFinished;
         sprintf(mess.text, "Unable to initialize the dynamic simulation.");
         if (dataCallbackFunc)
            dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
         return -1;
      }
   }

   sdstate(t, y, &y[sdm->nq]);

   if (init_muscle_states(sdm, y) == code_bad)
   {
      mess.id = dpFinished;
      sprintf(mess.text, "Unable to initialize the dynamic simulation.");
      if (dataCallbackFunc)
         dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
      return -1;
   }

   if (params.outputOptions.normalization != dpNoNormalization)
      params.stepSize = (params.endTime - params.startTime) / 99.0;

   if (output_kinetics_file)
      set_up_kinetics_output(output_kinetics_file);

   check_for_sderror("DLLMAIN");

#if CONTACT_DETECTION
   if (check_for_initial_penetration(y, dy) == code_bad)
      return 0;
#endif

   // Calculate initial accelerations.
   if (params.outputOptions.gencoordMoments.num_items > 0)
   {
      FILE* fp;
      dpBoolean newInverse = sdm->newInverseSimulation; // save a copy of this flag

      // For an inverse simulation, activate prescribed motion and prescribe
      // the Qs, Us, and As according to the input motion. Then write the
      // initial Qs and Us to a file. This is useful if, for example, you
      // later want to use the gencoord moments to drive a forward simulation
      // to reproduce the motion.
      if (newInverse == dpYes)
      {
         sdm->newInverseSimulation = dpNo;
         set_prescribed_motion(sdm, dpPrescribedQ, 1);
      }
      calc_derivatives(t, y, dy, param, &status);

      fp = fopen("initialConditions.txt","w");
		for (i = 0; i < sdm->nu; i++)
         fprintf(fp, "%s %21.16lf %21.16lf\n", sdm->q[i].name, y[i] * sdm->q[i].conversion, dy[i] * sdm->q[i].conversion);
      fclose(fp);

      // Set the newInverseSimulation flag back to what it was.
      sdm->newInverseSimulation = newInverse;
      if (newInverse == dpYes)
         set_prescribed_motion(sdm, dpPrescribedQ, 0);
   }
   else
   {
      calc_derivatives(t, y, dy, param, &status);
   }

   si.initial_system_energy = MINMDOUBLE;

   if (output_kinetics_file)
      write_kin_out_frame(t, y, dy);

   if (params.outputOptions.systemEnergy == dpYes)
      calc_system_energy(&si);

   if (params.outputOptions.jointReactionForces.num_items > 0 || params.outputOptions.jointReactionTorques.num_items > 0)
      get_reaction_forces();

   if (params.outputOptions.gencoordMoments.num_items > 0 ||
      (params.outputOptions.optimizedActivations_Muscles.num_items > 0 && params.outputOptions.optimizedActivations_Gencoords.num_items > 0) ||
      params.outputOptions.gencoordPowers.num_items > 0)
   {
      if (sdm->newInverseSimulation == dpYes)
         calc_initial_gencoord_moments(t, y, dy, input_data);
      else
         calc_gencoord_moments(t, y, dy);
   }

   // Moment arms should be calculated last because the function
   // calls sdstate() which resets all of the accelerations to zero.
   if ((params.outputOptions.muscleMomentArms_Muscles.num_items > 0 && params.outputOptions.muscleMomentArms_Gencoords.num_items > 0) ||
      (params.outputOptions.muscleMoments_Muscles.num_items > 0 && params.outputOptions.muscleMoments_Gencoords.num_items > 0) ||
      params.outputOptions.totalMuscleMoments.num_items > 0 ||
      (params.outputOptions.optimizedActivations_Muscles.num_items > 0 && params.outputOptions.optimizedActivations_Gencoords.num_items > 0))
      calc_muscle_moment_arms(t, y);

   if (params.outputOptions.optimizedActivations_Muscles.num_items > 0)
   {
      char message[255];

      if (set_up_optimization(message) != code_fine)
      {
         mess.id = dpFinished;
         sprintf(mess.text, "Could not set up muscle optimization: %s.", message);
         if (dataCallbackFunc)
            dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
         return -1;
      }
      optimize_muscle_activations(t, y);
   }

   if (params.outputMotionFile || dataCallbackFunc)
   {
      if (input_data && input_data->name)
      {
         // If the name of the input motion ends in "_IN", remove that
         // part and replace it with "_OUT".
         int len = strlen(input_data->name);
         strcpy(buffer, input_data->name);
         if (len >= 3 && STRINGS_ARE_EQUAL(&buffer[len - 3], "_IN"))
            strcpy(&buffer[len - 3], "_OUT");
      }
      else
      {
         strcpy(buffer, "sim_motion_OUT");
      }

      create_output_motion(buffer);
      if (params.outputMotionFile)
         open_output_motion_file(params.outputMotionFile);
   }

   // Now that the model is assembled, return a dpSetup frame to SIMM,
   // so it can create a motion to hold the simulation results.
   if (dataCallbackFunc)
   {
      dataSetup.startTime = params.startTime;
      dataSetup.endTime = params.endTime;

      get_motion_info(&dataSetup);

      dataCallbackFunc(sdm->simmModel, dpSetup, &dataSetup);

      dataFrame.numElements = dataSetup.numElements;
      dataFrame.elements = (double*)simm_malloc(dataFrame.numElements * sizeof(double));
   }

   // Store the starting position in the output motion, and return it to SIMM.
   if (dataCallbackFunc)
   {
      if (store_motion_frame(t, y, dy) != 0)
         sim_message(recover, "Error storing motion frame: %s", sd_error_msg);
      get_motion_data(&dataFrame);
      dataCallbackFunc(sdm->simmModel, dpFrame, &dataFrame);
   }
   else if (params.outputMotionFile)
   {
      if (store_motion_frame(t, y, dy) != 0)
         sim_message(recover, "Error storing motion frame: %s", sd_error_msg);
   }

   mess.id = dpInitialized;
   sprintf(mess.text, "Simulation (version %s) initialized at time = %lf", sim_version, t);
   if (dataCallbackFunc)
      dataCallbackFunc(sdm->simmModel, dpMessage, &mess);

   return 0;
}


/* dpRunSimulation: this function is called when the "run" button is pressed in
 * the Dynamics Tool. It contains the main integration for loop for the simulation.
 */
DLL int dpRunSimulation(int code)
{
   int i, err, which, status, num_impacts = 0;
   double step, step_size, next_t;
   double param[1] = {-1.0};
   dpMessageStruct mess;
   double start_time, end_time, elapsed;

   start_time = get_time_now();

   for (t = params.startTime; t < params.endTime - TINY_TIME; )
   {
      step = step_size = params.stepSize;
      next_t = t + step_size;
      if (next_t > params.endTime)
      {
         next_t = params.endTime;
         step = step_size = next_t - t;
      }

      // To make the simulation run faster, don't let the muscle excitations
      // change during an integration interval. Thus you can compute them here
      // and store them in the muscle structure. If you are spline-fitting
      // your excitation data, you don't really need to do this since the
      // muscle excitations will change smoothly (whereas with step functions
      // they will not), and thus they will not slow down the integration.
      // Note: Starting with version 2.0.1, all muscle derivative functions in
      // derivs.c depend on the muscle excitation being set here. If you remove
      // the code below, you will need to make sure that excitations are set
      // properly in the derivative function that you use.
      for (i = 0; i < sdm->num_muscles; i++)
         sdm->muscles[i].excitation_level = calc_excitation(&sdm->muscles[i], t, y);
      
      while (t < next_t - TINY_TIME)
      {
         calc_derivatives(t, y, dy, param, &status);
         if (status != NO_FLAG)
         {
            print_final_information(t, y, dy, &si);
            sim_message(none, "status != NO_FLAG.");            
            goto END_SIMULATION;
         }
         integrate(calc_derivatives, &t, y, dy, param, step_size, &step, sdm->neq, INTEGRATION_TOL, work, &err, &which);

         if (err == INT_NO_ERROR)
         {
            break;
         }
#if CONTACT_DETECTION
         else if (err == INT_USER_FLAG)
         {
            if (which == IMPACT_FLAG)
            {
               handle_impacts(t, y, dy);
               num_impacts++;
               step = step_size = next_t - t;
            }
            else if (which == CONS_VIOL)
            {
               store_motion_frame(t, y, dy);
               sim_message(abort_action, "CONSTRAINT VIOLATION at time %lf.", t);
               goto END_SIMULATION;
            }
            else if (which == BAD_STATE)
            {
               store_motion_frame(t, y, dy);
               sim_message(abort_action, "BAD STATE at time %lf.", t);
               goto END_SIMULATION;
            }
         } 
#endif
         else if (err == INT_OVER_BUMP)
         {
            if (which >= 0 && which < sdm->nu)
               sim_message(recover, "INTEGRATOR: went over a bump (gencoord %s [position]).", sdm->q[which].name);
            else if (which >= sdm->nu && which < sdm->nq + sdm->nu)
               sim_message(recover, "INTEGRATOR: went over a bump (gencoord %s [velocity]).", sdm->q[which - sdm->nq].name);
            else
               sim_message(recover, "INTEGRATOR: went over a bump (state vector element %d).", which);
            step_size = next_t - t;
         } 
         else if (err == INT_CANT_CONTINUE)
         {
            store_motion_frame(t, y, dy);
            print_final_information(t, y, dy, &si);
            sim_message(abort_action, "SDVINTEG: Cannot continue after time = %lf (gc %d).", t, which);
            goto END_SIMULATION;
         } 
         else 
         {
            print_final_information(t, y, dy, &si);
            if (which == CONS_VIOL)
               sim_message(abort_action, "Constraint Violation.");
            sim_message(abort_action, "%s", sd_error_msg);
            goto END_SIMULATION;
         }
      }

      assign_muscle_states(sdm, t, y);

      if (verbose == dpYes)
         print_state(y, dy);

      if (output_kinetics_file)
         write_kin_out_frame(t, y, dy);

      if (params.outputOptions.systemEnergy == dpYes)
         calc_system_energy(&si);

      if (params.outputOptions.jointReactionForces.num_items > 0 || params.outputOptions.jointReactionTorques.num_items > 0)
         get_reaction_forces();

      if (params.outputOptions.gencoordMoments.num_items > 0 ||
         (params.outputOptions.optimizedActivations_Muscles.num_items > 0 && params.outputOptions.optimizedActivations_Gencoords.num_items > 0) ||
         params.outputOptions.gencoordPowers.num_items > 0)
      {
         if (sdm->newInverseSimulation == dpYes)
         {
            if (t < params.endTime - TINY_TIME)
            {
               calc_next_gencoord_moments(t, y, dy, input_data);
               calc_corrective_moments(t, y, dy, sdm, input_data);
            }
            else
            {
               // You're at the end of the simulation, so you don't need to
               // calculate corrective moments for the next time step. Just get
               // the corrected moments from this time step and store them in
               // the sdm->q[] array so they can be written to the output motion
               // file and used by the static muscle optimization function.
					for (i = 0; i < sdm->nu; i++)
               {
                  if (input_data->q_moments[i])
                     sdm->q[i].moment = input_data->q_moments[i]->y[input_data->q_moments[i]->numpoints-1];
               }
            }
         }
         else
         {
            calc_gencoord_moments(t, y, dy);
         }
      }

      // Moment arms should be calculated near the end because the function
      // calls sdstate() which resets all of the accelerations to zero.
      if ((params.outputOptions.muscleMomentArms_Muscles.num_items > 0 && params.outputOptions.muscleMomentArms_Gencoords.num_items > 0) ||
         (params.outputOptions.muscleMoments_Muscles.num_items > 0 && params.outputOptions.muscleMoments_Gencoords.num_items > 0) ||
         params.outputOptions.totalMuscleMoments.num_items > 0 ||
         (params.outputOptions.optimizedActivations_Muscles.num_items > 0 && params.outputOptions.optimizedActivations_Gencoords.num_items > 0))
         calc_muscle_moment_arms(t, y);

      if (params.outputOptions.optimizedActivations_Muscles.num_items > 0 && params.outputOptions.optimizedActivations_Gencoords.num_items > 0)
         optimize_muscle_activations(t, y);

      if (dataCallbackFunc)
      {
         if (store_motion_frame(t, y, dy) != 0)
         {
            sim_message(recover, "Error storing motion frame: %s", sd_error_msg);
         }

         get_motion_data(&dataFrame);

         dataCallbackFunc(sdm->simmModel, dpFrame, &dataFrame);
      }
      else if (params.outputMotionFile)
      {
         store_motion_frame(t, y, dy);
      }

      mess.id = dpOverwritable;
      sprintf(mess.text, "SIM> Simulation currently at time = %lf", t);
      if (dataCallbackFunc)
         dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
   }

END_SIMULATION:
   if (params.outputMotionFile)
      close_output_motion_file();

   if (output_kinetics_file)
      fclose(kin_out);

   free(work);
   free(y);
   free(dy);

   if (params.outputOptions.optimizedActivations_Muscles.num_items > 0 && params.outputOptions.optimizedActivations_Gencoords.num_items > 0)
      cleanup_optimization();

   freeParameters(&params);

   FREE_IFNOTNULL(dataFrame.elements);
   FREE_IFNOTNULL(dataSetup.motionName);

   if (dataSetup.elementNames)
   {
      for (i = 0; i < dataSetup.numElements; i++)
         FREE_IFNOTNULL(dataSetup.elementNames[i]);
      FREE_IFNOTNULL(dataSetup.elementNames);
   }

   if (dataSetup.event)
   {
      for (i = 0; i < dataSetup.numEvents; i++)
         FREE_IFNOTNULL(dataSetup.event[i].name);
      FREE_IFNOTNULL(dataSetup.event);
   }

   if (gModelInfo)
   {
      FREE_IFNOTNULL(gModelInfo->joint_type);
      FREE_IFNOTNULL(gModelInfo);
   }

   mess.id = dpFinished;
   end_time = get_time_now();
   elapsed = (end_time - start_time)*3600;
   sprintf(mess.text, "Simulation finished at time = %lf, took %lfs", t, elapsed);
   if (dataCallbackFunc)
      dataCallbackFunc(sdm->simmModel, dpMessage, &mess);

   return 0;
}


/* dpPauseSimulation: this function is currently not used. When the "pause"
 * button is pressed in the Dynamics Tool, SIMM suspends the simulation thread,
 * so the simulation never receives a message that it has been paused.
 */
DLL int dpPauseSimulation(int code)
{
   return 0;
}


/* dpResetSimulation: this function is called when the "reset" button is pressed
 * in the Dynamics Tool. SIMM then frees the DLL from memory and exits the simulation
 * thread.
 */
DLL int dpResetSimulation(int code)
{
   if (sdm)
      FREE_IFNOTNULL(sdm->joint);

   return 0;
}


/* copyParameters: this function copies the simulation parameters into the DLL. 
 * The indices of the lists have not yet been set, this is done later in checkOutputOptions.
 */
void copyParameters(dpSimulationParameters* from, dpSimulationParameters* to)
{
   to->startTime = from->startTime;
   to->endTime = from->endTime;
   to->stepSize = from->stepSize;
   to->baumgarte_stabilization = from->baumgarte_stabilization;

   if (from->workingDir)
      mstrcpy(&to->workingDir, from->workingDir);

   if (from->inputMotionFile)
      mstrcpy(&to->inputMotionFile, from->inputMotionFile);

   if (from->outputMotionFile)
      mstrcpy(&to->outputMotionFile, from->outputMotionFile);

   if (from->boneDir)
      mstrcpy(&to->boneDir, from->boneDir);

   copy_name_list(&from->outputOptions.muscleActivations, &to->outputOptions.muscleActivations);
   copy_name_list(&from->outputOptions.muscleLengths, &to->outputOptions.muscleLengths);
   copy_name_list(&from->outputOptions.muscleFiberLengths, &to->outputOptions.muscleFiberLengths);
   copy_name_list(&from->outputOptions.muscleFiberVelocities, &to->outputOptions.muscleFiberVelocities);
   copy_name_list(&from->outputOptions.muscleForces, &to->outputOptions.muscleForces);
   to->outputOptions.contactForces = from->outputOptions.contactForces;
   copy_name_list(&from->outputOptions.massCenterPositions, &to->outputOptions.massCenterPositions);
   copy_name_list(&from->outputOptions.massCenterVelocities, &to->outputOptions.massCenterVelocities);
   to->outputOptions.systemEnergy = from->outputOptions.systemEnergy;
   copy_name_list(&from->outputOptions.jointReactionForces, &to->outputOptions.jointReactionForces);
   copy_name_list(&from->outputOptions.jointReactionTorques, &to->outputOptions.jointReactionTorques);
   copy_name_list(&from->outputOptions.gencoordMoments, &to->outputOptions.gencoordMoments);
   to->outputOptions.springForces = from->outputOptions.springForces;
   to->outputOptions.numSpringForces = from->outputOptions.numSpringForces;
   copy_name_list(&from->outputOptions.muscleMomentArms_Muscles, &to->outputOptions.muscleMomentArms_Muscles);
   copy_name_list(&from->outputOptions.muscleMomentArms_Gencoords, &to->outputOptions.muscleMomentArms_Gencoords);
   copy_name_list(&from->outputOptions.muscleMoments_Muscles, &to->outputOptions.muscleMoments_Muscles);
   copy_name_list(&from->outputOptions.muscleMoments_Gencoords, &to->outputOptions.muscleMoments_Gencoords);
   copy_name_list(&from->outputOptions.totalMuscleMoments, &to->outputOptions.totalMuscleMoments);
   copy_name_list(&from->outputOptions.optimizedActivations_Muscles, &to->outputOptions.optimizedActivations_Muscles);
   copy_name_list(&from->outputOptions.optimizedActivations_Gencoords, &to->outputOptions.optimizedActivations_Gencoords);
   to->outputOptions.normalization = from->outputOptions.normalization;
   copy_name_list(&from->outputOptions.gencoordValues, &to->outputOptions.gencoordValues);
   to->outputOptions.traditionalInverse = from->outputOptions.traditionalInverse;
   copy_name_list(&from->outputOptions.gencoordVelocities, &to->outputOptions.gencoordVelocities);
   copy_name_list(&from->outputOptions.gencoordAccelerations, &to->outputOptions.gencoordAccelerations);
   copy_name_list(&from->outputOptions.gencoordPowers, &to->outputOptions.gencoordPowers);
}


void freeParameters(dpSimulationParameters* params)
{
   FREE_IFNOTNULL(params->workingDir);
   FREE_IFNOTNULL(params->inputMotionFile);
   FREE_IFNOTNULL(params->outputMotionFile);
   FREE_IFNOTNULL(params->boneDir);

   allocate_name_list(&params->outputOptions.muscleActivations, 0);
   allocate_name_list(&params->outputOptions.muscleLengths, 0);
   allocate_name_list(&params->outputOptions.muscleFiberLengths, 0);
   allocate_name_list(&params->outputOptions.muscleFiberVelocities, 0);
   allocate_name_list(&params->outputOptions.muscleForces, 0);
   allocate_name_list(&params->outputOptions.massCenterPositions, 0);
   allocate_name_list(&params->outputOptions.massCenterVelocities, 0);
   allocate_name_list(&params->outputOptions.jointReactionForces, 0);
   allocate_name_list(&params->outputOptions.jointReactionTorques, 0);
   allocate_name_list(&params->outputOptions.gencoordMoments, 0);
   allocate_name_list(&params->outputOptions.muscleMomentArms_Muscles, 0);
   allocate_name_list(&params->outputOptions.muscleMomentArms_Gencoords, 0);
   allocate_name_list(&params->outputOptions.muscleMoments_Muscles, 0);
   allocate_name_list(&params->outputOptions.muscleMoments_Gencoords, 0);
   allocate_name_list(&params->outputOptions.totalMuscleMoments, 0);
   allocate_name_list(&params->outputOptions.optimizedActivations_Muscles, 0);
   allocate_name_list(&params->outputOptions.optimizedActivations_Gencoords, 0);
   allocate_name_list(&params->outputOptions.gencoordValues, 0);
   allocate_name_list(&params->outputOptions.gencoordVelocities, 0);
   allocate_name_list(&params->outputOptions.gencoordAccelerations, 0);
   allocate_name_list(&params->outputOptions.gencoordPowers, 0);
}


/* checkModel: this function checks to make sure that the SIMM model passed into
 * the DLL (via dpSetModel()) is the same model that is built into the DLL (iin
 * the file model.sd).
 */
int checkModel(dpModelStruct* model)
{
   int i, info[50], slider[6];
   dpBoolean fatalError = dpNo;

   mess.id = dpModelError;

   sdinfo(info);

   if (model->num_body_segments != (info[1] + 1))
   {
      if (dataCallbackFunc)
      {
         sprintf(mess.text, "SIM> Number of segments in simulation DLL (%d) does not match number in SIMM model (%d)",
            (info[1] + 1), model->num_body_segments);
         dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
      }
      fatalError = dpYes;
   }

   if (model->num_joints != info[1])
   {
      if (dataCallbackFunc)
      {
         sprintf(mess.text, "SIM> Number of tree joints in simulation DLL (%d) does not match number of joints in SIMM model (%d)",
            info[1], model->num_joints);
         dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
      }
      fatalError = dpYes;
   }

   if (model->num_closed_loops != info[4])
   {
      if (dataCallbackFunc)
      {
         sprintf(mess.text, "SIM> Number of loop joints in simulation DLL (%d) does not match number in SIMM model (%d)",
            info[4], model->num_closed_loops);
         dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
      }
      fatalError = dpYes;
   }

   if (fatalError == dpNo)
   {
      // Check only tree joints. If there are loop joints, model->joint and model->num_joints
      // will be increased and filled in later (in init_dll_model()).
      for (i = 0; i < model->num_joints; i++)
      {
         sdjnt(i, info, slider);
         switch (info[0])
         {
            case 1: // pin
               if (model->joint[i].jnt_type != dpPin && model->joint[i].jnt_type != dpReversePin && model->joint[i].jnt_type != dpWeld)
               {
                  if (dataCallbackFunc)
                  {
                     sprintf(mess.text, "SIM> Type of joint %d in simulation DLL (pin) does not match type in SIMM.", i + 1);
                     dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
                  }
                  fatalError = dpYes;
               }
               break;
            case 2: // ujoint
               if (model->joint[i].jnt_type != dpUniversal && model->joint[i].jnt_type != dpReverseUniversal)
               {
                  if (dataCallbackFunc)
                  {
                     sprintf(mess.text, "SIM> Type of joint %d in simulation DLL (universal) does not match type in SIMM.", i + 1);
                     dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
                  }
                  fatalError = dpYes;
               }
               break;
            case 3: // gimbal
               if (model->joint[i].jnt_type != dpGimbal && model->joint[i].jnt_type != dpReverseGimbal && model->joint[i].jnt_type != dpBall)
               {
                  if (dataCallbackFunc)
                  {
                     sprintf(mess.text, "SIM> Type of joint %d in simulation DLL (gimbal) does not match type in SIMM.", i + 1);
                     dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
                  }
                  fatalError = dpYes;
               }
               break;
            case 4: // ball
               if (model->joint[i].jnt_type != dpBall && model->joint[i].jnt_type != dpReverseBall && model->joint[i].jnt_type != dpGimbal)
               {
                  if (dataCallbackFunc)
                  {
                     sprintf(mess.text, "SIM> Type of joint %d in simulation DLL (ball) does not match type in SIMM.", i + 1);
                     dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
                  }
                  fatalError = dpYes;
               }
               break;
            case 5: // slider
               if (model->joint[i].jnt_type != dpSlider && model->joint[i].jnt_type != dpReverseSlider)
               {
                  if (dataCallbackFunc)
                  {
                     sprintf(mess.text, "SIM> Type of joint %d in simulation DLL (slider) does not match type in SIMM.", i + 1);
                     dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
                  }
                  fatalError = dpYes;
               }
               break;
            case 6: // free
               if (model->joint[i].jnt_type != dpFree && model->joint[i].jnt_type != dpBushing)
               {
                  if (dataCallbackFunc)
                  {
                     sprintf(mess.text, "SIM> Type of joint %d in simulation DLL (free) does not match type in SIMM.", i + 1);
                     dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
                  }
                  fatalError = dpYes;
                  if (model->joint[i].jnt_type == dpFree && model->joint[i].euler_rots == dpUndefinedOrder)
                  {
                     if (dataCallbackFunc)
                     {
                        sprintf(mess.text, "SIM> Quaternion joint %d of Free type has an undefined Euler order.", i + 1);
                        dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
                     }
                     fatalError = dpYes;
                  }
               }
               break;
            case 7: // cylindrical
               if (model->joint[i].jnt_type != dpCylindrical && model->joint[i].jnt_type != dpReverseCylindrical)
               {
                  if (dataCallbackFunc)
                  {
                     sprintf(mess.text, "SIM> Type of joint %d in simulation DLL (cylindrical) does not match type in SIMM.", i + 1);
                     dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
                  }
                  fatalError = dpYes;
               }
               break;
            case 8: // planar
               if (model->joint[i].jnt_type != dpPlanar)
               {
                  if (dataCallbackFunc)
                  {
                     sprintf(mess.text, "SIM> Type of joint %d in simulation DLL (planar) does not match type in SIMM.", i + 1);
                     dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
                  }
                  fatalError = dpYes;
               }
               break;
            case 9: // weld
               if (model->joint[i].jnt_type != dpWeld)
               {
                  if (dataCallbackFunc)
                  {
                     sprintf(mess.text, "SIM> Type of joint %d in simulation DLL (weld) does not match type in SIMM.", i + 1);
                     dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
                  }
                  fatalError = dpYes;
               }
               break;
            case 10: // bushing
               if (model->joint[i].jnt_type != dpBushing && model->joint[i].jnt_type != dpFree)
               {
                  if (dataCallbackFunc)
                  {
                     sprintf(mess.text, "SIM> Type of joint %d in simulation DLL (bushing) does not match type in SIMM.", i + 1);
                     dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
                  }
                  fatalError = dpYes;
               }
               break;
            case 11: // bearing
               if (model->joint[i].jnt_type != dpBearing)
               {
                  if (dataCallbackFunc)
                  {
                     sprintf(mess.text, "SIM> Type of joint %d in simulation DLL (bearing) does not match type in SIMM.", i + 1);
                     dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
                  }
                  fatalError = dpYes;
               }
               break;
            case 20: // rplanar
               if (model->joint[i].jnt_type != dpReversePlanar)
               {
                  if (dataCallbackFunc)
                  {
                     sprintf(mess.text, "SIM> Type of joint %d in simulation DLL (reverse planar) does not match type in SIMM.", i + 1);
                     dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
                  }
                  fatalError = dpYes;
               }
               break;
            case 21: // rfree
               if (model->joint[i].jnt_type != dpReverseFree)
               {
                  if (dataCallbackFunc)
                  {
                     sprintf(mess.text, "SIM> Type of joint %d in simulation DLL (reverse free) does not match type in SIMM.", i + 1);
                     dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
                  }
                  fatalError = dpYes;
               }
               break;
            case 22: // rbushing
               if (model->joint[i].jnt_type != dpReverseBushing)
               {
                  if (dataCallbackFunc)
                  {
                     sprintf(mess.text, "SIM> Type of joint %d in simulation DLL (reverse bushing) does not match type in SIMM.", i + 1);
                     dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
                  }
                  fatalError = dpYes;
               }
               break;
            case 23: // rbearing
               if (model->joint[i].jnt_type != dpReverseBearing)
               {
                  if (dataCallbackFunc)
                  {
                     sprintf(mess.text, "SIM> Type of joint %d in simulation DLL (reverse bearing) does not match type in SIMM.", i + 1);
                     dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
                  }
                  fatalError = dpYes;
               }
               break;
         }
      }
   }

   if (fatalError == dpYes)
      return -1;

   return 1;
}


/* This function checks all the specified simulation outputs, and determines the appropriate index
 * for each name. If a specified gencoord does not exist in the model, a warning is returned to SIMM.
 * Warnings for other missing elements are not returned because it is common for the set of muscles
 * to change, and extraneous bodies/joints are frequently removed (e.g., massless segments used only
 * for holding markers).
 */
int checkOutputOptions(void)
{
   int i;
   dpBoolean fatalError = dpNo;

   mess.id = dpModelWarning;

   // Set the indices for the output gencoord values.
   for (i=0; i<params.outputOptions.gencoordValues.num_items; i++)
   {
      if ((params.outputOptions.gencoordValues.index[i] = get_q_index_by_name(params.outputOptions.gencoordValues.name[i])) < 0)
      {
         if (dataCallbackFunc)
         {
            sprintf(mess.text, "SIM> Gencoord %s does not exist in the simulation DLL. Its value will not be output.",
               params.outputOptions.gencoordValues.name[i]);
            dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
         }
         remove_name_from_list(params.outputOptions.gencoordValues.name[i], &params.outputOptions.gencoordValues);
         i--;
      }
   }

   // Set the indices for the output gencoord velocities.
   for (i=0; i<params.outputOptions.gencoordVelocities.num_items; i++)
   {
      if ((params.outputOptions.gencoordVelocities.index[i] = get_q_index_by_name(params.outputOptions.gencoordVelocities.name[i])) < 0)
      {
         if (dataCallbackFunc)
         {
            sprintf(mess.text, "SIM> Gencoord %s does not exist in the simulation DLL. Its velocity will not be output.",
               params.outputOptions.gencoordVelocities.name[i]);
            dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
         }
         remove_name_from_list(params.outputOptions.gencoordVelocities.name[i], &params.outputOptions.gencoordVelocities);
         i--;
      }
   }

   // Set the indices for the output gencoord accelerations.
   for (i=0; i<params.outputOptions.gencoordAccelerations.num_items; i++)
   {
      if ((params.outputOptions.gencoordAccelerations.index[i] = get_q_index_by_name(params.outputOptions.gencoordAccelerations.name[i])) < 0)
      {
         if (dataCallbackFunc)
         {
            sprintf(mess.text, "SIM> Gencoord %s does not exist in the simulation DLL. Its acceleration will not be output.",
               params.outputOptions.gencoordAccelerations.name[i]);
            dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
         }
         remove_name_from_list(params.outputOptions.gencoordAccelerations.name[i], &params.outputOptions.gencoordAccelerations);
         i--;
      }
   }

   // Set the indices for the output muscle activations.
   for (i=0; i<params.outputOptions.muscleActivations.num_items; i++)
   {
      if ((params.outputOptions.muscleActivations.index[i] = get_muscle_index_by_name(params.outputOptions.muscleActivations.name[i])) < 0)
      {
         remove_name_from_list(params.outputOptions.muscleActivations.name[i], &params.outputOptions.muscleActivations);
         i--;
      }
   }

   // Set the indices for the output muscle lengths.
   for (i=0; i<params.outputOptions.muscleLengths.num_items; i++)
   {
      if ((params.outputOptions.muscleLengths.index[i] = get_muscle_index_by_name(params.outputOptions.muscleLengths.name[i])) < 0)
      {
         remove_name_from_list(params.outputOptions.muscleLengths.name[i], &params.outputOptions.muscleLengths);
         i--;
      }
   }

   // Set the indices for the output muscle forces.
   for (i=0; i<params.outputOptions.muscleForces.num_items; i++)
   {
      if ((params.outputOptions.muscleForces.index[i] = get_muscle_index_by_name(params.outputOptions.muscleForces.name[i])) < 0)
      {
         remove_name_from_list(params.outputOptions.muscleForces.name[i], &params.outputOptions.muscleForces);
         i--;
      }
   }

   // Set the indices for the output muscle fiber lengths.
   for (i=0; i<params.outputOptions.muscleFiberLengths.num_items; i++)
   {
      if ((params.outputOptions.muscleFiberLengths.index[i] = get_muscle_index_by_name(params.outputOptions.muscleFiberLengths.name[i])) < 0)
      {
         remove_name_from_list(params.outputOptions.muscleFiberLengths.name[i], &params.outputOptions.muscleFiberLengths);
         i--;
      }
   }

   // Set the indices for the output muscle fiber velocities.
   for (i=0; i<params.outputOptions.muscleFiberVelocities.num_items; i++)
   {
      if ((params.outputOptions.muscleFiberVelocities.index[i] = get_muscle_index_by_name(params.outputOptions.muscleFiberVelocities.name[i])) < 0)
      {
         remove_name_from_list(params.outputOptions.muscleFiberVelocities.name[i], &params.outputOptions.muscleFiberVelocities);
         i--;
      }
   }

   // Set the indices for the output mass center positions.  Store the index into the sdm->body_segment list, not the SD/FAST segment index.
   for (i=0; i<params.outputOptions.massCenterPositions.num_items; i++)
   {
      if ((params.outputOptions.massCenterPositions.index[i] = get_segment_index_by_name(params.outputOptions.massCenterPositions.name[i])) < 0)
      {
         remove_name_from_list(params.outputOptions.massCenterPositions.name[i], &params.outputOptions.massCenterPositions);
         i--;
      }
   }

   // Set the indices for the output mass center velocities.
   for (i=0; i<params.outputOptions.massCenterVelocities.num_items; i++)
   {
      if ((params.outputOptions.massCenterVelocities.index[i] = get_segment_index_by_name(params.outputOptions.massCenterVelocities.name[i])) < 0)
      {
         remove_name_from_list(params.outputOptions.massCenterVelocities.name[i], &params.outputOptions.massCenterVelocities);
         i--;
      }
   }

   // Set the indices for the output joint reaction forces.
   for (i=0; i<params.outputOptions.jointReactionForces.num_items; i++)
   {
      if ((params.outputOptions.jointReactionForces.index[i] = get_joint_index_by_name(params.outputOptions.jointReactionForces.name[i])) < 0)
      {
         remove_name_from_list(params.outputOptions.jointReactionForces.name[i], &params.outputOptions.jointReactionForces);
         i--;
      }
   }

   // Set the indices for the output joint reaction torques.
   for (i=0; i<params.outputOptions.jointReactionTorques.num_items; i++)
   {
      if ((params.outputOptions.jointReactionTorques.index[i] = get_joint_index_by_name(params.outputOptions.jointReactionTorques.name[i])) < 0)
      {
         remove_name_from_list(params.outputOptions.jointReactionTorques.name[i], &params.outputOptions.jointReactionTorques);
         i--;
      }
   }

   // Set the indices for the output gencoord moments.
   for (i=0; i<params.outputOptions.gencoordMoments.num_items; i++)
   {
      if ((params.outputOptions.gencoordMoments.index[i] = get_q_index_by_name(params.outputOptions.gencoordMoments.name[i])) < 0)
      {
         if (dataCallbackFunc)
         {
            sprintf(mess.text, "SIM> Gencoord %s does not exist in the simulation DLL. Its moment will not be output.",
               params.outputOptions.gencoordMoments.name[i]);
            dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
         }
         remove_name_from_list(params.outputOptions.gencoordMoments.name[i], &params.outputOptions.gencoordMoments);
         i--;
      }
   }

   // Set the indices for the output gencoord powers.
   for (i=0; i<params.outputOptions.gencoordPowers.num_items; i++)
   {
      if ((params.outputOptions.gencoordPowers.index[i] = get_q_index_by_name(params.outputOptions.gencoordPowers.name[i])) < 0)
      {
         if (dataCallbackFunc)
         {
            sprintf(mess.text, "SIM> Gencoord %s does not exist in the simulation DLL. Its power will not be output.",
               params.outputOptions.gencoordPowers.name[i]);
            dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
         }
         remove_name_from_list(params.outputOptions.gencoordPowers.name[i], &params.outputOptions.gencoordPowers);
         i--;
      }
   }

   // Set the indices for the output moment arms (muscles).
   for (i=0; i<params.outputOptions.muscleMomentArms_Muscles.num_items; i++)
   {
      if ((params.outputOptions.muscleMomentArms_Muscles.index[i] = get_muscle_index_by_name(params.outputOptions.muscleMomentArms_Muscles.name[i])) < 0)
      {
         remove_name_from_list(params.outputOptions.muscleMomentArms_Muscles.name[i], &params.outputOptions.muscleMomentArms_Muscles);
         i--;
      }
   }

   // Set the indices for the output moment arms (gencoords).
   for (i=0; i<params.outputOptions.muscleMomentArms_Gencoords.num_items; i++)
   {
      if ((params.outputOptions.muscleMomentArms_Gencoords.index[i] = get_q_index_by_name(params.outputOptions.muscleMomentArms_Gencoords.name[i])) < 0)
      {
         if (dataCallbackFunc)
         {
            sprintf(mess.text, "SIM> Gencoord %s does not exist in the simulation DLL. Moment arms for this gencoord will not be output.",
               params.outputOptions.muscleMomentArms_Gencoords.name[i]);
            dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
         }
         remove_name_from_list(params.outputOptions.muscleMomentArms_Gencoords.name[i], &params.outputOptions.muscleMomentArms_Gencoords);
         i--;
      }
   }

   // Set the indices for the output muscle moments (muscles).
   for (i=0; i<params.outputOptions.muscleMoments_Muscles.num_items; i++)
   {
      if ((params.outputOptions.muscleMoments_Muscles.index[i] = get_muscle_index_by_name(params.outputOptions.muscleMoments_Muscles.name[i])) < 0)
      {
         remove_name_from_list(params.outputOptions.muscleMoments_Muscles.name[i], &params.outputOptions.muscleMoments_Muscles);
         i--;
      }
   }

   // Set the indices for the output muscle moments (gencoords).
   for (i=0; i<params.outputOptions.muscleMoments_Gencoords.num_items; i++)
   {
      if ((params.outputOptions.muscleMoments_Gencoords.index[i] = get_q_index_by_name(params.outputOptions.muscleMoments_Gencoords.name[i])) < 0)
      {
         if (dataCallbackFunc)
         {
            sprintf(mess.text, "SIM> Gencoord %s does not exist in the simulation DLL. Muscle moments for this gencoord will not be output.",
               params.outputOptions.muscleMoments_Gencoords.name[i]);
            dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
         }
         remove_name_from_list(params.outputOptions.muscleMoments_Gencoords.name[i], &params.outputOptions.muscleMoments_Gencoords);
         i--;
      }
   }

   // Set the indices for the output total muscle moments.
   for (i=0; i<params.outputOptions.totalMuscleMoments.num_items; i++)
   {
      if ((params.outputOptions.totalMuscleMoments.index[i] = get_q_index_by_name(params.outputOptions.totalMuscleMoments.name[i])) < 0)
      {
         if (dataCallbackFunc)
         {
            sprintf(mess.text, "SIM> Gencoord %s does not exist in the simulation DLL. Total muscle moments for this gencoord will not be output.",
               params.outputOptions.totalMuscleMoments.name[i]);
            dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
         }
         remove_name_from_list(params.outputOptions.totalMuscleMoments.name[i], &params.outputOptions.totalMuscleMoments);
         i--;
      }
   }

   // Set the indices for the output optimized activations (muscles).
   for (i=0; i<params.outputOptions.optimizedActivations_Muscles.num_items; i++)
   {
      if ((params.outputOptions.optimizedActivations_Muscles.index[i] = get_muscle_index_by_name(params.outputOptions.optimizedActivations_Muscles.name[i])) < 0)
      {
         remove_name_from_list(params.outputOptions.optimizedActivations_Muscles.name[i], &params.outputOptions.optimizedActivations_Muscles);
         i--;
      }
   }

   // Set the indices for the output optimized activations (gencoords).
   for (i=0; i<params.outputOptions.optimizedActivations_Gencoords.num_items; i++)
   {
      if ((params.outputOptions.optimizedActivations_Gencoords.index[i] = get_q_index_by_name(params.outputOptions.optimizedActivations_Gencoords.name[i])) < 0)
      {
         if (dataCallbackFunc)
         {
            sprintf(mess.text, "SIM> Gencoord %s does not exist in the simulation DLL. Muscle activations will not be optimized for this gencoord.",
               params.outputOptions.optimizedActivations_Gencoords.name[i]);
            dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
         }
         remove_name_from_list(params.outputOptions.optimizedActivations_Gencoords.name[i], &params.outputOptions.optimizedActivations_Gencoords);
         i--;
      }
   }

   // Make sure each muscle specifies a muscle model, and if static optimization is chosen, set the model accordingly.
   for (i=0; i<sdm->num_muscles; i++)
   {
      if (find_name_in_list(sdm->muscles[i].name, &params.outputOptions.optimizedActivations_Muscles) >= 0)
      {
         // This muscle's activation is being optimized. Set its model to muscle_model_opt_index.
         if (sdm->muscles[i].muscle_model_opt_index == NULL)
         {
            if (dataCallbackFunc)
            {
               sprintf(mess.text, "SIM> muscle_model_opt is undefined for muscle %s.", sdm->muscles[i].name);
               dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
            }
            fatalError = dpYes;
            break;
         }
         else
         {
            if (sdm->muscles[i].muscle_model_index == NULL || sdm->muscles[i].muscle_model_index == sdm->default_muscle.muscle_model_index)
               sdm->muscles[i].muscle_model_index = (int*)simm_calloc(1, sizeof(int));
            *sdm->muscles[i].muscle_model_index = *sdm->muscles[i].muscle_model_opt_index;
         }
      }
      else
      {
         if (sdm->muscles[i].muscle_model_index == NULL)
         {
            if (dataCallbackFunc)
            {
               sprintf(mess.text, "SIM> muscle_model is undefined for muscle %s.", sdm->muscles[i].name);
               dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
            }
            fatalError = dpYes;
            break;
         }
      }
   }

   if (fatalError == dpYes)
      return -1;

   return 1;
}


/* copyModel: this function copies one dpModelStruct to another. It is used
 * to copy the model in SIMM memory to one in DLL memory.
 */
void copyModel(dpModelStruct* from, dpModelStruct* to)
{
   int i, j;

   memcpy(to, from, sizeof(dpModelStruct));

   if (from->name)
      mstrcpy(&to->name, from->name);

#if CONTACT_DETECTION
   init_contact_info();
#else
   sdm->num_contacts = sdm->contacts_size = 0;
   sdm->num_bilat_contacts = sdm->bilat_contacts_size = 0;
   sdm->contacts = NULL;
   sdm->bilat_contacts = NULL;
#endif

   if (to->num_springs > 0 && from->spring)
   {
      to->spring = (dpSpringStruct*)simm_malloc(to->num_springs * sizeof(dpSpringStruct));
      memcpy(to->spring, from->spring, to->num_springs * sizeof(dpSpringStruct));
      for (i = 0; i < to->num_springs; i++)
         mstrcpy(&to->spring[i].contact_name, from->spring[i].contact_name);
      to->spring_array_size = to->num_springs;
   }

   if (to->num_spring_floors > 0 && from->spring_floor)
   {
      to->spring_floor = (dpSpringFloor*)simm_malloc(to->num_spring_floors * sizeof(dpSpringFloor));
      memcpy(to->spring_floor, from->spring_floor, to->num_spring_floors * sizeof(dpSpringFloor));
      for (i = 0; i < to->num_spring_floors; i++)
      {
         mstrcpy(&to->spring_floor[i].name, from->spring_floor[i].name);
         if (from->spring_floor[i].ph)
         {
            to->spring_floor[i].ph = (dpPolyhedronStruct*)simm_malloc(sizeof(dpPolyhedronStruct));
            copyPolyhedron(from->spring_floor[i].ph, to->spring_floor[i].ph);
         }
      }
   }

   if (to->num_spring_surfaces > 0 && from->spring_surface)
   {
      to->spring_surface = (dpSpringSurface*)simm_malloc(to->num_spring_surfaces * sizeof(dpSpringSurface));
      memcpy(to->spring_surface, from->spring_surface, to->num_spring_surfaces * sizeof(dpSpringSurface));
      for (i = 0; i < to->num_spring_surfaces; i++)
      {
         mstrcpy(&to->spring_surface[i].name, from->spring_surface[i].name);
         if (from->spring_surface[i].ph)
         {
            to->spring_surface[i].ph = (dpPolyhedronStruct*)simm_malloc(sizeof(dpPolyhedronStruct));
            copyPolyhedron(from->spring_surface[i].ph, to->spring_surface[i].ph);
         }
      }
   }

   if (to->num_force_mattes > 0 && from->force_matte)
   {
      to->force_matte = (dpForceMatte*)simm_malloc(to->num_force_mattes * sizeof(dpForceMatte));
      for (i = 0; i < to->num_force_mattes; i++)
      {
         mstrcpy(&to->force_matte[i].name, from->force_matte[i].name);
         to->force_matte[i].segment = from->force_matte[i].segment;
         if (from->force_matte[i].ph)
         {
            to->force_matte[i].ph = (dpPolyhedronStruct*)simm_malloc(sizeof(dpPolyhedronStruct));
            copyPolyhedron(from->force_matte[i].ph, to->force_matte[i].ph);
         }
      }
   }

   if (to->num_body_segments > 0 && from->body_segment)
   {
      to->body_segment = (dpBodyStruct*)simm_malloc(to->num_body_segments * sizeof(dpBodyStruct));
      memcpy(to->body_segment, from->body_segment, to->num_body_segments * sizeof(dpBodyStruct));
      for (i = 0; i < to->num_body_segments; i++)
      {
         dpBodyStruct* body = &to->body_segment[i];
         mstrcpy(&body->name, from->body_segment[i].name);
         body->contact_joints = NULL;
         if (body->num_objects > 0 && from->body_segment[i].object)
         {
            body->object = (dpPolyhedronStruct**)simm_malloc(body->num_objects * sizeof(dpPolyhedronStruct*));
            for (j = 0; j < body->num_objects; j++)
            {
               body->object[j] = (dpPolyhedronStruct*)simm_malloc(sizeof(dpPolyhedronStruct));
               copyPolyhedron(from->body_segment[i].object[j], body->object[j]);
            }
         }
         else
         {
            body->object = NULL;
         }
      }
   }

   copyMuscles(from, to);

   /* Most of the joint array is filled in later, but some of it needs to be copied
    * so you can check to make sure the models match.
    */
   if (to->num_joints > 0)
   {
      to->joint = (dpJointStruct*)simm_malloc(to->num_joints * sizeof(dpJointStruct));
      memcpy(to->joint, from->joint, to->num_joints * sizeof(dpJointStruct));
   }

   if (to->num_wrap_objects > 0 && from->wrap_object)
   {
      to->wrap_object = (dpWrapObject*)simm_malloc(to->num_wrap_objects * sizeof(dpWrapObject));
      memcpy(to->wrap_object, from->wrap_object, to->num_wrap_objects * sizeof(dpWrapObject));
      for (i = 0; i < to->num_wrap_objects; i++)
         mstrcpy(&to->wrap_object[i].name, from->wrap_object[i].name);
   }

   if (to->num_constraint_objects > 0 && from->constraint_object)
   {
      to->constraint_object = (dpConstraintObject*)simm_malloc(to->num_constraint_objects * sizeof(dpConstraintObject));
      memcpy(to->constraint_object, from->constraint_object, to->num_constraint_objects * sizeof(dpConstraintObject));
      for (i = 0; i < to->num_constraint_objects; i++)
      {
         dpConstraintObject* obj = &to->constraint_object[i];
         mstrcpy(&obj->name, from->constraint_object[i].name);
         if (obj->numPoints > 0 && from->constraint_object[i].points)
         {
            obj->points = (dpConstraintPoint*)simm_malloc(obj->numPoints * sizeof(dpConstraintPoint));
            memcpy(obj->points, from->constraint_object[i].points, obj->numPoints * sizeof(dpConstraintPoint));
            for (j = 0; j < obj->numPoints; j++)
               mstrcpy(&obj->points[j].name, from->constraint_object[i].points[j].name);
         }
      }
   }

   /* copy all the constraint, restraint and muscle point functions */
   if (to->num_functions > 0 && from->function)
   {
      to->function = (dpFunction**)simm_malloc(to->num_functions * sizeof(dpFunction*));
      for (i = 0; i < to->num_functions; i++)
      {
         to->function[i] = (dpFunction *)simm_calloc(1, sizeof(dpFunction));
         malloc_function(to->function[i], from->function[i]->numpoints);
         copyFunction(from->function[i], to->function[i]);
      }
   }

   /* This must be done after the constraint functions because the Qs contain
    * pointers to the functions.
    */
   if (to->nu > 0 && from->q)
   {
      to->q = (dpQStruct*)simm_malloc(to->nu * sizeof(dpQStruct));
      memcpy(to->q, from->q, to->nu * sizeof(dpQStruct));
      for (i = 0; i < to->nu; i++)
      {
         mstrcpy(&to->q[i].name, from->q[i].name);
         to->q[i].restraint_function = getFunctionPointer(to, from, from->q[i].restraint_function);
         to->q[i].min_restraint_function = getFunctionPointer(to, from, from->q[i].min_restraint_function);
         to->q[i].max_restraint_function = getFunctionPointer(to, from, from->q[i].max_restraint_function);
         to->q[i].constraint_function = getFunctionPointer(to, from, from->q[i].constraint_function);
      }
   }

   to->enforce_constraints = from->enforce_constraints;
}


/* getFunctionPointer: this function scans an array of dpSplineFunctions (the 'from' array),
 * looking for one that matches the passed-in pointer. It returns a pointer to the
 * corresponding dpFunction in the 'to' array.
 */
dpFunction* getFunctionPointer(dpModelStruct* to, dpModelStruct* from, dpFunction* fromPtr)
{
   if (to && from && fromPtr)
   {
      int i;

      for (i = 0; i < from->num_functions; i++)
         if (fromPtr == from->function[i])
            return to->function[i];
   }

   return NULL;
}


/* COPYFUNCTION: copies a function structure. This structure contains
 * the original (x,y) data points and the coefficients (b,c,d) of
 * the function interpolants. It assumes that space for the x, y, a, b, c
 * arrays has already been allocated, and the size of this space
 * is stored in to->coefficient_array_size.
 */
void copyFunction(dpFunction* from, dpFunction* to)
{
   int numToCopy = MIN(to->coefficient_array_size, from->coefficient_array_size);

   to->type = from->type;
   to->numpoints = from->numpoints;
   to->source = from->source;
   to->status = from->status;
   to->used = from->used;
   to->usernum = from->usernum;
   to->cutoff_frequency = from->cutoff_frequency;

   memcpy(to->x, from->x, numToCopy * sizeof(double));
   memcpy(to->y, from->y, numToCopy * sizeof(double));
   memcpy(to->b, from->b, numToCopy * sizeof(double));
   memcpy(to->c, from->c, numToCopy * sizeof(double));
   memcpy(to->d, from->d, numToCopy * sizeof(double));
}

/* copyPolyhedron: this function makes a copy of a dpPolyhedronStruct. */
void copyPolyhedron(dpPolyhedronStruct* from, dpPolyhedronStruct* to)
{
   int i;

   if (from->name)
      mstrcpy(&to->name, from->name);

   to->body_segment = from->body_segment;

   memcpy(&to->bc, &from->bc, sizeof(dpBoundingCube));

   if (from->num_vertices > 0 && from->vertex)
   {
      to->num_vertices = from->num_vertices;
      to->vertex = (dpVertexStruct*)simm_malloc(to->num_vertices * sizeof(dpVertexStruct));
      memcpy(to->vertex, from->vertex, to->num_vertices * sizeof(dpVertexStruct));
      for (i = 0; i < to->num_vertices; i++)
      {
         to->vertex[i].polygons = (int*)simm_malloc(to->vertex[i].polygon_count * sizeof(int));
         memcpy(to->vertex[i].polygons, from->vertex[i].polygons, to->vertex[i].polygon_count * sizeof(int));
      }
   }
   else
   {
      to->num_vertices = 0;
      to->vertex = NULL;
   }

   if (from->num_polygons > 0 && from->polygon)
   {
      to->num_polygons = from->num_polygons;
      to->polygon = (dpPolygonStruct*)simm_malloc(to->num_polygons * sizeof(dpPolygonStruct));
      memcpy(to->polygon, from->polygon, to->num_polygons * sizeof(dpPolygonStruct));
      for (i = 0; i < to->num_polygons; i++)
      {
         to->polygon[i].vertex_index = (int*)simm_malloc(to->polygon[i].num_vertices * sizeof(int));
         memcpy(to->polygon[i].vertex_index, from->polygon[i].vertex_index, to->polygon[i].num_vertices * sizeof(int));
      }
   }
   else
   {
      to->num_polygons = 0;
      to->polygon = NULL;
   }
}


/* copyMusles: this function copies the default muscle and array of muscles
 * from one dpModelStruct (in SIMM memory) to another (in DLL memory).
 */
void copyMuscles(dpModelStruct* from, dpModelStruct* to)
{
   int i;

   to->num_muscles = from->num_muscles;

   to->muscles = (dpMuscleStruct*)simm_malloc(to->num_muscles * sizeof(dpMuscleStruct));

   copyDefaultMuscle(&from->default_muscle, &to->default_muscle, from, to);

   for (i = 0; i < to->num_muscles; i++)
      copyMuscle(&from->muscles[i], &from->default_muscle, &to->muscles[i], &to->default_muscle, from, to);
}


/* copyDefaultMuscle: this function makes a copy of the default muscle.
 * Don't handle points in default muscle any more */
void copyDefaultMuscle(dpMuscleStruct* from, dpMuscleStruct* to, dpModelStruct* ms, dpModelStruct* dp)
{
   int i;

   if (from->name == NULL)
      to->name = NULL;
   else
      mstrcpy(&to->name, from->name);

   to->nummomentarms = 0;
   to->momentarms = NULL;

   copy_nndouble(from->max_isometric_force, &to->max_isometric_force);
   copy_nndouble(from->pennation_angle, &to->pennation_angle);
   copy_nndouble(from->max_contraction_vel, &to->max_contraction_vel);
   copy_nndouble(from->optimal_fiber_length,&to->optimal_fiber_length);
   copy_nndouble(from->resting_tendon_length,&to->resting_tendon_length);

   to->num_dynamic_params = from->num_dynamic_params;

   if (to->num_dynamic_params > 0)
   {
      int i;

      to->dynamic_param_names = (char**)simm_malloc(to->num_dynamic_params * sizeof(char*));

      for (i=0; i<to->num_dynamic_params; i++)
         mstrcpy(&to->dynamic_param_names[i], from->dynamic_param_names[i]);

      to->dynamic_params = (double**)simm_malloc(to->num_dynamic_params * sizeof(double*));

      for (i = 0; i < to->num_dynamic_params; i++)
         copy_nndouble(from->dynamic_params[i], &to->dynamic_params[i]);
   }
   else
   {
      to->dynamic_param_names = NULL;
      to->dynamic_params = NULL;
   }

   if (from->active_force_len_func == NULL)
      to->active_force_len_func = NULL;
   else
   {
      to->active_force_len_func = (dpFunction**)simm_malloc(sizeof(dpFunction*));
      *to->active_force_len_func = *from->active_force_len_func;
   }

   if (from->passive_force_len_func == NULL)
      to->passive_force_len_func = NULL;
   else
   {
      to->passive_force_len_func = (dpFunction**)simm_malloc(sizeof(dpFunction*));
      *to->passive_force_len_func = *from->passive_force_len_func;
   }

   if (from->tendon_force_len_func == NULL)
      to->tendon_force_len_func = NULL;
   else
   {
      to->tendon_force_len_func = (dpFunction**)simm_malloc(sizeof(dpFunction*));
      *to->tendon_force_len_func = *from->tendon_force_len_func;
   }

   if (from->force_vel_func == NULL)
      to->force_vel_func = NULL;
   else
   {
      to->force_vel_func = (dpFunction**)simm_malloc(sizeof(dpFunction*));
      *to->force_vel_func = *from->force_vel_func;
   }

   if (from->excitation_abscissa == NULL)
      to->excitation_abscissa = NULL;
   else
   {
      to->excitation_abscissa = (void**)simm_malloc(sizeof(void*));
      *to->excitation_abscissa = *from->excitation_abscissa;
   }

   if (from->excitation_func == NULL)
      to->excitation_func = NULL;
   else
   {
      to->excitation_func = (dpFunction**)simm_malloc(sizeof(dpFunction*));
      *to->excitation_func = *from->excitation_func;
   }

   copy_nnint(from->muscle_model_index, &to->muscle_model_index);
   copy_nnint(from->muscle_model_opt_index, &to->muscle_model_opt_index);
}


/* sim_message: this function returns a message and error status to SIMM.
 * There is a separate version of it for the stand-alone version of a simulation.
 */
int sim_message(ErrorAction action, const char* format, ...)
{
   int n = 0;
   va_list ap;
   char buffer3[CHARBUFFER];
   static dpMessageStruct mess;

   if (dataCallbackFunc)
   {
      va_start(ap, format);
      n = vsprintf(buffer3, format, ap);
      va_end(ap);

      strcpy(mess.text, "SIM> ");
      strncat(mess.text, buffer3, sizeof(mess.text) - 6);

      if (action == none)
         mess.id = dpNormal;
      else if (action == recover)
         mess.id = dpSimWarning;
      else if (action == abort_action)
         mess.id = dpSimError;
      else if (action == initialized)
         mess.id = dpInitialized;
      else if (action == overwrite)
         mess.id = dpOverwritable;
      else
         mess.id = dpFinished;

      dataCallbackFunc(sdm->simmModel, dpMessage, &mess);
   }

   return n;
}
