/*******************************************************************************

   FUNCTIONS.H

   Authors: Peter Loan
            Krystyne Blaikie

   Copyright (c) 1996-2011 MusculoGraphics, a division of Motion Analysis Corp.
   All rights reserved.

   Description: This file contains function prototypes for all functions which
      are callable by the GMC or SD/FAST user.

*******************************************************************************/

#ifndef FUNCTIONS_H
#define FUNCTIONS_H

int    acpp_main(int argc, char* argv[]);
void   append_if_necessary (char* str, char c);
void   apply_external_forces(double time, double y[], MotionData* data);
void   apply_gencoord_restraint_moments(dpModelStruct* sdm, double state_quat[]);
void   apply_muscle_forces(dpModelStruct* sdm);
ReturnCode assemble_model(double t, double y[]);
void   assign_muscle_states(dpModelStruct* sdm, double time, double state_quat[]);
double calc_active_force(dpMuscleStruct* muscle, double norm_fiber_length);
double calc_distance_between_points(double p1[], int n1, double p2[], int n2);
double calc_excitation(dpMuscleStruct* muscle, double time, double state_quat[]);
double calc_fiber_force(dpMuscleStruct* muscle, double activation, double norm_fiber_length, double norm_fiber_velocity);
double calc_fiber_velocity(dpMuscleStruct* muscle, double activation, double active_force, double velocity_dependent_force);
double calc_force_velocity(dpMuscleStruct* muscle, double norm_fiber_velocity);
double calc_distance_to_ellipsoid(double p1[], double radii[], double projpt[]);//for constraint objects
void   calc_gencoord_moments(double time, double state_quat[], double dstate_quat[]);
void   calc_ligament_force(dpMuscleStruct* muscle, double state_euler[]);
int    calc_line_intersect_cylinder (
	double		p11[3],		/* input:  muscle point 1 */
	double		p22[3],		/* input:  muscle point 2 */
   double		p0[3],      /* input:  point on cylinder axis */
	double		dn[3],      /* input:  direction vector of cyl axis */
	double		r,          /* input:  cylinder radius */
	double		len,        /* input:  cylinder length */
	double*		rlen,		   /* output: length of surface wrap */
	double		r1[],		   /* output: wrap tangent point 1 */
	double		r2[],		   /* output: wrap tangent point 2 */
	double**	wrap_pts,	   /* output: intermediate surface wrap pts */
	int*		num_wrap_pts,	/* output: number of intermediate pts */
	int		wrap_axis,	   /* input:  constraint axis */
	int		wrap_sign,	   /* input:  constraint direction */
	dpMuscleWrapStruct* mwrap, /* in/out: muscle wrap structure */
	int*		p_flag,		   /* output: did wrapping occur? */
	dpWrapObject*	wo);        /* input:  pointer to the cylinder wrap object */
int    calc_line_intersect_ellipsoid (
	double		p1[3],		/* input:  muscle point 1		*/
	double		p2[3],		/* input:  muscle point 2		*/
	double		m[3],		/* input:  ellipsoid origin (0,0,0)	*/
	double		a[3],		/* input:  ellipsoid xyz radii		*/
	double*		rlen,		/* output: length of surface wrap	*/
	double		r1[],		/* output: wrap tangent point 1		*/
	double		r2[],		/* output: wrap tangent point 2		*/
	double**	wrap_pts,	/* output: intermediate surface wrap pts*/
	int*		num_wrap_pts,	/* output: number of intermediate pts	*/
	int		wrap_axis,	/* input:  constraint axis		*/
	int		wrap_sign,	/* input:  constraint direction		*/
	dpMuscleWrapStruct* mwrap, /* in/out: muscle wrap structure	*/
	int*		p_flag,		/* output: did wrapping occur?		*/
	dpWrapObject*	wo);
int    calc_line_intersect_sphere(double p1[], double p2[], double m[],
			       double r, double *rlen, double r1[],
			       double r2[],
			       double** wrap_pts, int *num_wrap_pts,
			       int wrap_axis, int wrap_sign, dpMuscleWrapStruct* mwrap,
			       int *p_flag,
			       dpWrapObject* wo);
int    calc_line_intersect_torus (
	double		p1[3],		/* input:  muscle point 1		*/
	double		p2[3],		/* input:  muscle point 2		*/
	double		m[3],		/* input:  torus origin (0,0,0)	*/
	double		a[3],		/* input:  torus xy[z] radii		*/
	double*		rlen,		/* output: length of surface wrap	*/
	double		r1[],		/* output: wrap tangent point 1		*/
	double		r2[],		/* output: wrap tangent point 2		*/
	double**	wrap_pts,	/* output: intermediate surface wrap pts*/
	int*		num_wrap_pts,	/* output: number of intermediate pts	*/
	int		wrap_axis,	/* input:  constraint axis		*/
	int		wrap_sign,	/* input:  constraint direction		*/
	dpMuscleWrapStruct* mwrap, /* in/out: muscle wrap structure	*/
	int*		p_flag,		/* output: did wrapping occur?		*/
	dpWrapObject*	wo);
void   calc_muscle_power(dpMuscleStruct* muscle);
void   calc_muscle_tendon_force(dpMuscleStruct* muscle, double activation, double state_euler[]);
double calc_muscle_tendon_length(dpMuscleStruct* muscle, double state_euler[]);
double calc_muscle_tendon_velocity(dpMuscleStruct* muscle);
double calc_nonzero_passive_force(dpMuscleStruct* muscle, double norm_fiber_length, double norm_fiber_velocity);
double calc_passive_force(dpMuscleStruct* muscle, double norm_fiber_length, double norm_fiber_velocity);
double calc_pennation(double fiber_length, double optimal_fiber_length, double initial_pennation_angle);
void   calc_system_energy(SystemInfo* si);
double calc_tendon_force(dpMuscleStruct* muscle, double norm_tendon_length);
#if CONTACT_DETECTION
ReturnCode   check_object_info(void);
#endif
void   check_wrapping_points(dpMuscleStruct* muscle, double state_euler[]);
char*  clean_string(char buf[]);
void   convert_point(double pt[], int frame1, int frame2);
void   convert_state_from_quaternion(double time, double q[], double u[], double udot[], double q_out[], double u_out[], double udot_out[], double previous_state[]);
ReturnCode   convert_state_to_quaternion(double time, double q[], double u[], double udot[], double q_out[], double u_out[], double udot_out[]);
void   convert_string(char str[]);
void   convert_moment_array_from_quaternion(double state_euler[], double input_moments[], double output_moments[]);
void   convert_moment_array_to_quaternion(double state_quat[], double input_moments[], double output_moments[]);
void   convert_moment_to_quaternion(double theta1, double theta2, double theta3, double input_moments[], double output_moments[], dpEulerAngleOrder order);
int    count_muscle_states(dpModelStruct* sdm);
dpQStruct *enter_gencoord(dpModelStruct *model, char gencoord_name[]);
int    enter_segment(dpModelStruct* sdm, char segment_name[]);
int    enter_function(dpModelStruct *model, int usernum, dpBoolean permission_to_add);
void   error(ErrorAction action, char str_buffer[]);
void   evaluate_active_movingpoints(dpMuscleStruct *muscle, double state_euler[]);
void   evaluate_orig_movingpoints(dpMuscleStruct *muscle, double state_euler[]);
void   free_function(dpFunction* func);
void   free_motion_data(MotionData* md);
void   get_applied_force(double time, MotionData* data, int frc_index, double appl_point[], double force_vec[], double free_torque[]);
void   get_applied_torque(double time, MotionData* data, int trq_index, double torque_vec[]);
const  char* get_function_suffix(dpFunction* func);
ReturnCode get_muscle_file_name(char params_file[], char** muscle_file);
int    get_muscle_param_index(dpMuscleStruct* muscle, char param_name[]);
double get_muscle_param_value(dpMuscleStruct* muscle, char param_name[]);
ReturnCode get_string_pair(char str_buffer[], char str1[], char str2[]);
const  char* get_suffix(const char str[]);
void   get_temp_file_name(const char* fileName, char tempFileName[]);
dpWrapObject* get_wrap_object(char name[]);
ReturnCode get_xypair_from_string(char str_buffer[], double* x, double* y);
int    identify_elements_to_transfer(MotionData* data, dpBoolean transfer[]);
void   init_constraint_objects(void);
ReturnCode init_functions(dpModelStruct* model);
void   init_joint_functions(void);
void   init_joint_names(void);
void   init_motion(void);
ReturnCode init_muscle_states(dpModelStruct* sdm, double state_quat[]);
#if CONTACT_DETECTION
void   init_object_info(void);
#endif
void   init_q_restraint_functions(void);
void   init_qs(void);
void   init_quaternion_joints(void);
void   init_segments(void);
ReturnCode init_function(dpFunction* sf, double Ycoords[], int numpoints, dpFunctionType type,
                            double cutoffFrequency);
void   init_wrap_objects(void);
dpBoolean is_absolute_path(char *path);
dpBoolean is_name_in_list(const char* name, dpNameList* list);
dpBoolean is_preference_on(const char value[]);
void lmdif_C(void (*fcn)(int, int, double[], double[], int *, void *),
             int m, int n, double x[], double fvec[], double ftol, double xtol,
             double gtol, int maxfev, double epsfcn, double diag[], int mode,
             double factor, int nprint, int *info, int *nfev, double fjac[],
             int ldfjac, int ipvt[], double qtf[], double wa1[], double wa2[],
             double wa3[], double wa4[], void *fcn_data);
ReturnCode load_input_data(dpModelStruct* sdm, MotionData* data, char filename[]);
dpFunction* load_simm_function(dpModelStruct* model, dpFunction* func);
dpFunction* load_user_function(dpModelStruct* model, dpFunction* func);
ReturnCode make_zero_func(dpFunction* sf, double min_x, double max_x);
ReturnCode malloc_function(dpFunction* func, int numpoints);
ReturnCode mstrcpy(char* dest_str[], const char original_str[]);
int name_is_muscle(char name[], dpModelStruct* sdm, char suffix[], dpFunctionType* functionType, double* cutoffFrequency);
char*  parse_string(char str_buffer[], VariableType var_type, void* dest_var);
FILE*  preprocess_file(char infile[], char outfile[]);
#if CONTACT_DETECTION
ReturnCode read_contact_pair(FILE* fp);
ReturnCode read_contact_parameters_file(char params_file[]);
#endif
ReturnCode read_dllparameters_file(char params_file[], dpModelStruct* sdm);
ReturnCode read_double_array(FILE* fp, const char ending[], char name[], dpFunction* func);
ReturnCode read_kinetics_file(dpModelStruct* sdm, MotionData* data, char filename[]);
int    read_line(FILE* fp, char str_buffer[]);
ReturnCode read_motion_file(dpModelStruct* sdm, MotionData* data, char filename[]);
ReturnCode read_muscle_attachment_points(FILE* fp, dpModelStruct *sdm, dpMuscleStruct *muscle);
ReturnCode read_muscle_file(dpModelStruct* sdm, char filename[], dpIncludeMuscles include);
ReturnCode read_muscle_groups(FILE* fp);
int    read_nonempty_line(FILE* fp, char str_buffer[]);
#if CONTACT_DETECTION
ReturnCode read_object(FILE* fp);
ReturnCode read_object_group(FILE* fp);
#endif
ReturnCode read_parameters_file(char params_file[], dpModelStruct* sdm, char **muscle_file, char **input_data_file, 
                                char** output_motion_file, char** output_kinetics_file);
ReturnCode read_polyhedron(dpPolyhedronStruct* ph, char filename[]);
int    read_string(FILE* fp, char str_buffer[]);
int    read_til_tokens(FILE* file, char buf[], const char delimiters[]);
ReturnCode realloc_function(dpFunction* func, int size);
int    sduforce(double t, double q[], double u[]);
int    sdumotion(double t, double q[], double u[]);
void   set_fixed_gencoords(dpModelStruct* sdm);
ReturnCode   set_initial_conditions(double *t, double y[], double dy[], MotionData *input_motion);
void   set_prescribed_gencoords(double time, double q_quat[], double u_quat[], dpModelStruct* sdm, MotionData* data);
void   set_prescribed_motion(dpModelStruct* sdm, dpQType type, int value);
void   set_up_input_data(char filename[], MotionData** data);
void   set_up_kinetics_output(char filename[]);
ReturnCode setup_functions(dpModelStruct* sdm, MotionData* data);
FILE*  simm_fopen (const char* name, const char* mode);
void*  simm_malloc(unsigned mem_size);
void*  simm_calloc(unsigned num_elements, unsigned elem_size);
void*  simm_realloc(void* ptr, unsigned mem_size, ReturnCode* rc);
int    sim_message(ErrorAction action, const char* format, ...);
void strip_brackets_from_string(char name[]);
ReturnCode verify_assembly(dpModelStruct* sdm, double state_quat[]);
void   write_kin_out_frame(double time, double y[], double dy[]);

/* Functions defined in readtools.c */
int _read_til_tokens(FILE* file, char* buf, const char delimiters[]);
void _strip_outer_whitespace(char* str);

/* Function Prototypes for the user-created muscle-state functions */

/* Functions defined in INITS.C */
int muscle_init_func1(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[]);
int muscle_init_func2(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[]);
int muscle_init_func3(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[]);
int muscle_init_func4(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[]);
int muscle_init_func5(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[]);
int muscle_init_func6(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[]);
int muscle_init_func7(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[]);
int muscle_init_func8(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[]);
int muscle_init_func9(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[]);
int muscle_init_func10(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[]);
int muscle_init_func11(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[]);
int muscle_init_func12(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[]);
int muscle_init_func13(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[]);
int muscle_init_func14(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[]);
int muscle_init_func15(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[]);
int muscle_init_func16(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[]);
int muscle_init_func17(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[]);
int muscle_init_func18(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[]);
int muscle_init_func19(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[]);
int muscle_init_func20(dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[]);

/* functions defined in DERIVS.C */
int muscle_deriv_func1(double time, dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[],
		       double dstate_quat[], double *muscle_force);
int muscle_deriv_func2(double time, dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[],
		       double dstate_quat[], double *muscle_force);
int muscle_deriv_func3(double time, dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[],
		       double dstate_quat[], double *muscle_force);
int muscle_deriv_func4(double time, dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[],
		       double dstate_quat[], double *muscle_force);
int muscle_deriv_func5(double time, dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[],
		       double dstate_quat[], double *muscle_force);
int muscle_deriv_func6(double time, dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[],
		       double dstate_quat[], double *muscle_force);
int muscle_deriv_func7(double time, dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[],
		       double dstate_quat[], double *muscle_force);
int muscle_deriv_func8(double time, dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[],
		       double dstate_quat[], double *muscle_force);
int muscle_deriv_func9(double time, dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[],
		       double dstate_quat[], double *muscle_force);
int muscle_deriv_func10(double time, dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[],
			double dstate_quat[], double *muscle_force);
int muscle_deriv_func11(double time, dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[],
			double dstate_quat[], double *muscle_force);
int muscle_deriv_func12(double time, dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[],
			double dstate_quat[], double *muscle_force);
int muscle_deriv_func13(double time, dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[],
			double dstate_quat[], double *muscle_force);
int muscle_deriv_func14(double time, dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[],
			double dstate_quat[], double *muscle_force);
int muscle_deriv_func15(double time, dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[],
			double dstate_quat[], double *muscle_force);
int muscle_deriv_func16(double time, dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[],
			double dstate_quat[], double *muscle_force);
int muscle_deriv_func17(double time, dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[],
			double dstate_quat[], double *muscle_force);
int muscle_deriv_func18(double time, dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[],
			double dstate_quat[], double *muscle_force);
int muscle_deriv_func19(double time, dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[],
			double dstate_quat[], double *muscle_force);
int muscle_deriv_func20(double time, dpMuscleStruct *muscle, int muscle_state_start_index, double state_euler[],
			double dstate_quat[], double *muscle_force);

/* functions defined in ASSIGNS.C */
int muscle_assign_func1(double time, dpMuscleStruct *muscle, double state_euler[]);
int muscle_assign_func2(double time, dpMuscleStruct *muscle, double state_euler[]);
int muscle_assign_func3(double time, dpMuscleStruct *muscle, double state_euler[]);
int muscle_assign_func4(double time, dpMuscleStruct *muscle, double state_euler[]);
int muscle_assign_func5(double time, dpMuscleStruct *muscle, double state_euler[]);
int muscle_assign_func6(double time, dpMuscleStruct *muscle, double state_euler[]);
int muscle_assign_func7(double time, dpMuscleStruct *muscle, double state_euler[]);
int muscle_assign_func8(double time, dpMuscleStruct *muscle, double state_euler[]);
int muscle_assign_func9(double time, dpMuscleStruct *muscle, double state_euler[]);
int muscle_assign_func10(double time, dpMuscleStruct *muscle, double state_euler[]);
int muscle_assign_func11(double time, dpMuscleStruct *muscle, double state_euler[]);
int muscle_assign_func12(double time, dpMuscleStruct *muscle, double state_euler[]);
int muscle_assign_func13(double time, dpMuscleStruct *muscle, double state_euler[]);
int muscle_assign_func14(double time, dpMuscleStruct *muscle, double state_euler[]);
int muscle_assign_func15(double time, dpMuscleStruct *muscle, double state_euler[]);
int muscle_assign_func16(double time, dpMuscleStruct *muscle, double state_euler[]);
int muscle_assign_func17(double time, dpMuscleStruct *muscle, double state_euler[]);
int muscle_assign_func18(double time, dpMuscleStruct *muscle, double state_euler[]);
int muscle_assign_func19(double time, dpMuscleStruct *muscle, double state_euler[]);
int muscle_assign_func20(double time, dpMuscleStruct *muscle, double state_euler[]);

/* functions defined in PIPETOOLS.C */
void add_moments_to_motion_data(dpModelStruct* sdm, MotionData* data);
void calc_corrective_moments(double time, double q[], double u[], dpModelStruct* sdm, MotionData* data);
void calc_initial_gencoord_moments(double time, double y[], double dy[], MotionData* data);
void calc_next_gencoord_moments(double time, double y[], double dy[], MotionData* data);
int  check_for_sderror(char caller[]);
void convert_point2(int frame1, double p1[], int frame2, double p2[]);
void init_joints(dpBoolean allocateSpace);
ReturnCode convert_angles_to_quaternions(double q[], double u[], double udot[], double q_out[], double u_out[], double udot_out[], dpEulerAngleOrder order, int axes[]);
int get_q_index_by_name(const char name[]);
int get_segment_index_by_name(const char name[]);
int get_joint_index_by_name(const char name[]);
int get_muscle_index_by_name(const char name[]);
void allocate_name_list(dpNameList* list, int size);
void clear_name_list(dpNameList* list);
int add_name_to_list(const char* name, dpNameList* list);
void remove_name_from_list(const char* name, dpNameList* list);
int find_name_in_list(const char* name, dpNameList* list);
void copy_name_list(dpNameList* from, dpNameList* to);

/* functions defined in INVERT.C */
void ludcmp(double* a[], int n, int indx[], double* d);
void lubksb(double* a[], int n, int indx[], double b[]);

/* functions defined in LCP_SOLVER.C */
#if CONTACT_DETECTION
void compute_contact_forces(int impact_flag);
#endif

/* functions defined in IMPACT.C */
#if CONTACT_DETECTION
void handle_impacts(double t, double y[], double dy[]);
#endif

/* functions defined in CONTACT.C */
#if CONTACT_DETECTION
void apply_forces_at_contacts(int n, dpContactInfo list[]);
ReturnCode check_for_initial_penetration(double y[], double dy[]);
void copy_cnode(dpContactInfo from, dpContactInfo *to);
void create_bilateral_contacts(double y[]);
void determine_contacts(double y[], double dy[]);
void get_bilat_contact_info(void);
void get_contact_info(void);
double get_norm_rel_vel(int i, double vn[]);
double get_rel_vel(int cont, double rel_vel[]);
double get_tang_rel_vel(int i, double vt[]);
void init_contact_info(void);
ReturnCode makepaths(void);
void print_contact_list(dpContactInfo list[], int n);
#endif

/* functions defined in GMC.C */
void calc_muscle_derivatives(double time, double state_quat[], double dstate_quat[], double param[], int* status);

/* functions defined in DLLMAIN.C */
void init_dll_model(void);
int checkModel(dpModelStruct* model);
int checkOutputOptions(void);
void copyParameters(dpSimulationParameters* from, dpSimulationParameters* to);
void freeParameters(dpSimulationParameters* params);
void copyModel(dpModelStruct* from, dpModelStruct* to);
dpFunction* getFunctionPointer(dpModelStruct* to, dpModelStruct* from, dpFunction* fromPtr);
void copyFunction(dpFunction* from, dpFunction* to);
void copyPolyhedron(dpPolyhedronStruct* from, dpPolyhedronStruct* to);
void copyMuscles(dpModelStruct* from, dpModelStruct* to);
void copyDefaultMuscle(dpMuscleStruct* from, dpMuscleStruct* to, dpModelStruct* ms, dpModelStruct* dp);
void copyMuscle(dpMuscleStruct* from, dpMuscleStruct* defFrom, dpMuscleStruct* to, dpMuscleStruct* defTo,
                dpModelStruct* ms, dpModelStruct* dp);

/* functions defined in MAIN.C */
void calc_derivatives(double t, double y[], double dy[], double param[], int *status);

/* functions defined in OUTPUT.c */
void get_reaction_forces(void);
double get_time_now(void);
void print_body_info(void);
void print_final_information(double time, double y[], double dy[], SystemInfo* si);
void print_simulation_info(double time, double state_quat[], SystemInfo* si);
void print_state(double y[], double dy[]);
void show_array(char name[], double vect[], int start, int end);
void show_array_as_matrix(char name[], double vec[], int n, int m);
void show_matrix(char name[], double *mat[], int m, int n);
void show_vector(char name[], double vect[]);
void time_elapsed(double t_in, double t_out);
void create_output_motion(const char name[]);
void open_output_motion_file(const char name[]);
int store_motion_frame(double time, double state_quat[], double dstate_quat[]);
void get_motion_info(dpDataSetup* dataSetup);
void get_motion_data(dpDataFrame* dataFrame);

/* functions defined in OBJECT.c */
void copy_polyhedron(dpPolyhedronStruct* from, dpPolyhedronStruct* to);
void post_process_polyhedron(dpPolyhedronStruct *poly, double scale[]);
#if CONTACT_DETECTION
void show_object_info(void);
#endif

/* spring-based contact detection routines defined in spring_based.c */
int get_spring_contact_segment_sd_index(int springIndex);
void apply_spring_forces(dpModelStruct *sdm);
void show_spring_info(void);
dpBoolean vector_intersects_matte(double pt[], double vec[], dpForceMatte* fm, double inter[]);

/* functions defined in constraint.c */
double calculate_constraint_position_error(dpConstraintObject *co, int status, int point);
double calculate_constraint_velocity_error(dpConstraintObject *co, int point);
double calculate_constraint_acceleration_error(dpConstraintObject *co, int point);
void   determine_projected_point(dpConstraintObject *co, double pt[], double projpt[], double norm[]);
void apply_constraint_forces(dpConstraintObject *co, double multiplier, int j);

/* functions defined in actoptimization.c */
ReturnCode set_up_optimization(char message[]);
void cleanup_optimization(void);
void optimize_muscle_activations(double time, double state[]);
int count_optimized_gencoords(void);
void append_optimized_gencoord_residuals(MotionData* md, int* column, int frame);

#endif /* FUNCTIONS_H */
