/*******************************************************************************

   MATHTOOLS.H

   Author: Peter Loan

   Copyright (c) 1998-2010 MusculoGraphics, a division of Motion Analysis Corp.
   All rights reserved.

*******************************************************************************/

/* functions defined in mathtools.c */
void     append_matrix(DMatrix m, DMatrix b);
ReturnCode calc_function_coefficients(dpFunction* sf);
void     clear_vector(double a[], int n);
double   compute_angle_between_vectors(double vector1[], double vector2[]);
void     convert_rotation_to_quaternion(double R[][3], double q[]);
void     copy_3x3matrix(double from[][3], double to[][3]);
void     copy_4x4matrix(double from[][4], double to[][4]);
void     copy_vector(double from[], double to[], int length);
void     cross_vectors(double vector1[], double vector2[], double result[]);
double   distance_between_vertices(double vertex1[], double vertex2[]);
double   distancesqr_between_vertices(double vertex1[], double vertex2[]);
void     extract_euler_angles(double R[][4], double xyz_rot[3], dpEulerAngleOrder order);
int      find_end_of_array(int list[], int* length);    
void     format_double(double number, char format[]);
double   get_distsqr_point_line(double point[], double pl[], double vl[]);
void     get_point_from_point_line(double point[], double pt[], double vec[],
                               double closest_pt[]);
void     integrate(void (*func)(), double *time, double st[], double dst[], double param[],
                 double dt, double *step, int neqin, double tol, double work[], int *err, int *which);
double   interpolate_function(double abscissa, dpFunction* func, Derivative deriv,
                          double velocity, double acceleration);
dpBoolean intersect_line_plane01(double pt1[], double pt2[], 
                                 double plane[], double d, double inter[], double* t);
dpBoolean intersect_line_plane(double pt1[], double pt2[], 
                               double plane[], double d, double inter[], double* t);
int      intersect_lines(double p1[], double p2[], double p3[], double p4[],
                    double p_int1[], double* t, double p_int2[], double* s);
int      intersect_lines_scaled(double p1[], double p2[], double p3[], double p4[],
                           double p_int1[], double* t, double* mag1,
                           double p_int2[], double* s, double* mag2);

void     make_3x3dircos_matrix(double angle, double axis[], double mat[][3]);     
void     make_3x3matrix(double m[3][3], double m11, double m12, double m13, 
                    double m21, double m22, double m23, double m31, double m32, double m33);
void     make_3x3_xrot_matrix(double a, double m[][3]);
void     make_4x4dircos_matrix(double angle, double axis[], double mat[][4]);
void     make_vector(double vec[3], double x, double y, double z);
void     mult_3x3matrices(double mat1[][3], double mat2[][3], double result[][3]);
void     mult_3x3matrix_by_vector(double mat[][3], double vector[], double result[]);
void     mult_4x4matrices(double mat1[][4], double mat2[][4], double result[][4]);    
void     mult_4x4matrix_by_vector(double mat[][4], double vector[], double result[]);

double   normalize_vector(double vector[], double norm_vector[]);
int      point_in_polygon2D(double point[], double pgon[][2], int numverts);
dpBoolean point_in_polygon3D(double pt[], dpPolyhedronStruct* ph, int polygon_index); 
void     print_3x3matrix(double mat[][3], char string[], FILE* stream);
double   project_point_onto_plane(double pt[], dpPlaneStruct* plane, double projpt[]);
void     reset_3x3matrix(double matrix[][3]);
void     reset_4x4matrix(double mat[][4]);
void     rotate_matrix_axis_angle(double m[][4], const double* axis, double angle);
void     set_rotation_matrix_from_quaternion(double q[], double R[][4]);
void     set_rotation_matrix_from_euler_angles(double angle1, int axis1, double angle2, int axis2, 
                                         double angle3, int axis3, double R[][3], dpEulerAngleOrder order);
void     transform_pt(double m[][4], double* pt);
void     transform_vec(double m[][4], double* vec);
void     translate_matrix(double m[][4], double* delta);
void     transpose_3x3matrix(double mat[][3], double mat_transpose[][3]);
void     transpose_4x4matrix(double mat[][4], double mat_transpose[][4]);
dpBoolean vect_equal_within_tol(double a[3], double b[3], double tol);
dpBoolean zero_3x3matrix(double m[][3]);

void x_rotate_matrix_bodyfixed(double m[][4], double angle);
void y_rotate_matrix_bodyfixed(double m[][4], double angle);
void z_rotate_matrix_bodyfixed(double m[][4], double angle);
void x_rotate_matrix_spacefixed(double m[][4], double angle);
void y_rotate_matrix_spacefixed(double m[][4], double angle);
void z_rotate_matrix_spacefixed(double m[][4], double angle);


/* defined in invert.c */
void     invert_3x3matrix(double matrix[][3], double inverse[][3]);
void     invert_4x4matrix(double matrix[][4], double inverse[][4]);
void     invert_4x4transform(double matrix[][4], double inverse[][4]);
void     invert_matrix(double* matrix[], double* inverse[], int size);



