/*******************************************************************************

   STRUCTS.H

   Authors: Peter Loan
            Krystyne Blaikie

   Copyright (c) 1996-2011 MusculoGraphics, a division of Motion Analysis Corp.
   All rights reserved.

   Description: This file contains definitions of all structures that are
      used by the Dynamics Pipeline.

*******************************************************************************/

#ifndef STRUCTS_H
#define STRUCTS_H


STRUCT {
   int segment;                      /* body segment to which force is applied */
   dpFunction appl_point[3];         /* point of application of force (local coords) */
   dpFunction force_vec[3];          /* force vector */
   dpFunction free_torque[3];        /* torque vector (for ground-reaction forces only) */
} ForceStruct;                       /* contains one external force */


STRUCT {
   int segment;                      /* body segment to which torque is applied */
   dpFunction torque_vec[3];         /* torque vector (in coords of body segment) */
} TorqueStruct;                      /* contains one external torque */


STRUCT {
   double pointOfApplication[3];
   double force[3];
   double freeTorque[3];
   double magnitude;
} ForceMatteForce;


STRUCT {
   char* name;                         /* name of this motion */
   FILE* fp;                           /* file that these data read from or written to */
   int num_frames;                     /* number of [time] frames of data */
   int num_otherdata;                  /* number of "other" data values per frame */
   int num_elements;                   /* number of data values per frame */
   int num_musc_excitations;           /* number of muscle excitations */
   int num_forces;                     /* number of external forces */
   int num_torques;                    /* number of external torques */
   int current_frame;                  /* current data frame in simulation */
   double range_start;                 /* start of X range */
   double range_end;                   /* end of X range */
   double step_size;                   /* step size between adjacent frames of data */
   char* units;                        /* units of motion range */
   char** elementnames;                /* names of the data elements */
   double** motiondata;                /* the actual motion data */
   dpFunction** q_data;                /* pointers to the q (gencoord) data */
   dpFunction** u_data;                /* pointers to the u (gencoord vels) data */
   dpFunction** udot_data;             /* pointers to the udot (gencoord accels) */
   dpFunction** q_moments;             /* pointers to the moments applied to the qs */
   dpFunction** musc_excitations;      /* pointers to the muscle excitations */
   ForceStruct** forces;               /* external forces (e.g. ground-reaction) */
   TorqueStruct** torques;             /* external torques (e.g. ground-reaction) */
   int num_events;                     /* number of events in array */
   dpMotionEvent* event;               /* array of motion events (e.g., heelstrike) */
   float event_color[3];               /* color used to display events in SIMM plots */
   int time_column;                    /* column index of time data, if any */
   int x_column;                       /* index of column that serves as X variable in SIMM */
   dpBoolean enforce_constraints;      /* enforce constraints while prescribing motion? */
   dpBoolean enforce_loops;            /* not used in simulation, just copied to output motion */
   dpBoolean show_markers;             /* not used in simulation, just copied to output motion */
   dpBoolean show_marker_trails;       /* not used in simulation, just copied to output motion */
   dpBoolean show_forces;              /* not used in simulation, just copied to output motion */
   dpBoolean show_force_trails;        /* not used in simulation, just copied to output motion */
   dpBoolean show_foot_prints;         /* not used in simulation, just copied to output motion */
   dpBoolean show_force_plates;        /* not used in simulation, just copied to output motion */
   dpBoolean show_gencoord_moments;    /* not used in simulation, just copied to output motion */
   dpBoolean show_other_objects;       /* not used in simulation, just copied to output motion */
   dpBoolean wrap;                     /* not used in simulation, just copied to output motion */
   int num_foot_prints;                /* number of foot prints */
   dpFootPrint** foot_print;           /* array of pointers to foot prints */
   int num_force_plates;               /* number of force plates */
   dpForcePlate** force_plate;         /* array of pointers to force plates */
} MotionData;                          /* holds motion data for inverse dynamics */



STRUCT {
   double mass;
   double gravity[3];
   double mass_center[3];
   double inertia[3][3];
   double kinetic_energy;
   double potential_energy;
   double linear_momentum[3];
   double angular_momentum[3];
   double system_energy;
   double initial_system_energy;
} SystemInfo;


#endif /* STRUCTS_H */
